import unittest
import tempfile
from pathlib import Path
import json

from ibllib.misc import version
from oneibl import one, registration

one = one.ONE(base_url='https://test.alyx.internationalbrainlab.org', username='test_user',
              password='TapetesBloc18')

md5_0 = 'add2ab27dbf8428f8140-0870d5080c7f'
r = {'created_by': 'olivier',
     'path': 'clns0730/2018-08-24/002',
     'filenames': ["raw_behavior_data/_iblrig_encoderTrialInfo.raw.ssv"],
     'hashes': [md5_0],
     'filesizes': [1234],
     'versions': [version.ibllib()]}


class TestRegistrationEndpoint(unittest.TestCase):

    def test_single_registration(self):
        dataset = one.alyx.rest('register-file', 'create', data=r)
        ds = one.alyx.rest('datasets', 'read', id=dataset[0]['id'])
        self.assertEqual(ds['hash'], md5_0)
        self.assertEqual(ds['file_size'], 1234)
        self.assertEqual(ds['version'], version.ibllib())
        self.assertEqual(len(dataset[0]['file_records']), 2)
        one.alyx.rest('datasets', 'delete', id=dataset[0]['id'])
        # self.assertEqual(ds['collection'], 'raw_behavior_data')

    def test_registration_server_only(self):
        # in this case the repository has only one file record
        r_ = r.copy()
        r_['server_only'] = True
        # r_['name'] = 'ibl_patcher'
        dataset = one.alyx.rest('register-file', 'create', data=r_)[0]
        self.assertEqual(len(dataset['file_records']), 1)
        self.assertEqual(dataset['file_records'][0]['data_repository'], 'flatiron_mainenlab')
        one.alyx.rest('datasets', 'delete', id=dataset['id'])


class TestRegistrationSession(unittest.TestCase):

    def test_registration_session(self):
        settings = {
            'SESSION_DATE': '2018-04-01',
            'SESSION_DATETIME': '2018-04-01T12:48:26.795526',
            'PYBPOD_CREATOR': ['test_user',
                               'f092c2d5-c98a-45a1-be7c-df05f129a93c',
                               'local'],
            'SESSION_NUMBER': '002',
            'SUBJECT_NAME': 'clns0730',
            'PYBPOD_BOARD': '_iblrig_mainenlab_behavior_1',
            'PYBPOD_PROTOCOL': '_iblrig_tasks_ephysChoiceWorld',
            'IBLRIG_VERSION_TAG': '5.4.1',
            'SUBJECT_WEIGHT': 22,
        }
        with tempfile.TemporaryDirectory() as td:
            # creates the local session
            session_path = Path(td).joinpath('clns0730', '2018-04-01', '002')
            alf_path = session_path.joinpath('alf')
            alf_path.mkdir(parents=True)
            alf_path.joinpath('spikes.times.npy').touch()
            alf_path.joinpath('spikes.amps.npy').touch()
            behavior_path = session_path.joinpath('raw_behavior_data')
            behavior_path.mkdir()
            settings_file = behavior_path.joinpath('_iblrig_taskSettings.raw.json')
            with open(settings_file, 'w') as fid:
                json.dump(settings, fid)
            rc = registration.RegistrationClient(one=one)
            rc.register_session(session_path)
            eid = one.search(subjects='clns0730', date_range=['2018-04-01', '2018-04-01'])[0]
            datasets = one.alyx.get('/datasets?subject=clns0730&date=2018-04-01')
            for ds in datasets:
                self.assertTrue(ds['hash'] is not None)
                self.assertTrue(ds['file_size'] is not None)
                self.assertTrue(ds['version'] == version.ibllib())
            one.alyx.rest('sessions', 'delete', id=eid)



class SyncBpodWheelException(ValueError):
    """
    When the bpod can't be synchronized with the data from Bonsai
    """
    pass


class SyncBpodFpgaException(ValueError):
    """
    When the bpod can't be synchronized with the FPGA because of overlap issue
    """
    pass



"""IBL pipeline.

A task is defined by:

- a name (also name of a Python function)
- a list of relative paths (possibly with a *) to input files/directories (relative to the
  session directory)
- a list of relative paths (possibly with a *) to output files/directories (relative to the
  session directory)
- a list of dependencies (other task names)
- n_cpus
- n_gpus
- io_charge (integer between 0 and 100)
- priority

A scheduler (possibly based on dask) accepts a list of tasks and executes them
on the available resources (CPUs, GPUs).

The list of required tasks is obtained by scanning the existing directory for
tasks that need to be run.

"""

# -------------------------------------------------------------------------------------------------
# Imports
# -------------------------------------------------------------------------------------------------

from datetime import datetime
from itertools import chain
import logging
from pathlib import Path
import time

from dask import visualize
from dask.distributed import Client, LocalCluster


logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------------------------------
# Utils
# -------------------------------------------------------------------------------------------------

def get_full_paths(session_dir, path_templates):
    """Return a list of paths matching the given templates in the session directory."""
    return list(chain(*(session_dir.glob(tmp) for tmp in path_templates)))


def all_file_globs_exist(session_dir, path_templates):
    """Return whether all path templates (with a wildcard *) exist or not."""
    return (
        len(path_templates) > 0 and
        0 not in [len(list(session_dir.glob(p))) for p in path_templates])


STATUS_LIST = ('pending', 'processing', 'completed', 'error')


def is_session_dir(path):
    """Return whether a directory is a session directory or not."""
    # TODO: improve how we determine whether a dir is a session dir or not.
    return (Path(path) / 'raw_ephys_data').exists()


def find_session_dirs(roots):
    """Recursively find a list of session dirs in one or several root directories."""
    for r in roots:
        for subdir in Path(r).rglob('*'):
            if subdir.is_dir() and is_session_dir(subdir):
                yield subdir


# -------------------------------------------------------------------------------------------------
# Base task class
# -------------------------------------------------------------------------------------------------

class Task:
    """Base class for the tasks."""

    # To set in the child classes.
    name = None
    input_path_templates = ()
    output_path_templates = ()
    depends_on = ()

    # Charge.
    n_cpus = 0
    n_gpus = 0
    io_charge = 0
    priority = None

    # Task status.
    # See also # https://distributed.dask.org/en/latest/scheduling-state.html#task-state
    _status = 'pending'
    started = None
    completed = None

    def __init__(self, session_dir, **kwargs):
        self.__dict__.update(kwargs)
        self.session_dir = Path(session_dir).resolve()
        self.input_paths = get_full_paths(self.session_dir, self.input_path_templates)
        if self.is_complete():
            self.status = 'completed'

    def run(self):
        """To override."""
        print("run", self)
        time.sleep(10)
        #
        # for i in range(10):
        # return np.exp(np.random.normal(int(1e2)))

    @property
    def status(self):
        """Task status: one of %s.""" % ', '.join(STATUS_LIST)
        return self._status

    @status.setter
    def status(self, status):
        assert status in STATUS_LIST
        self._status = status

    def all_input_files_exist(self):
        """Return whether all input files exist so that the task can run."""
        return all_file_globs_exist(self.session_dir, self.input_path_templates)

    def all_output_files_exist(self):
        """Return whether all output files have been created."""
        return all_file_globs_exist(self.session_dir, self.output_path_templates)

    def is_complete(self):
        """Return whether the task is complete, by default, if all output files have been created.
        This method can be overriden for custom checks that the task has successfully completed."""
        return self.all_output_files_exist()

    def __repr__(self):
        return '<Task %s (%s) in %s>' % (self.name, self.status, self.session_dir)


# -------------------------------------------------------------------------------------------------
# IBL tasks
# -------------------------------------------------------------------------------------------------

class ExtractBehaviorTask(Task):
    name = 'extract_behaviour'
    output_path_templates = ('alf/_ibl_trials.*',)
    n_cpus = 1
    io_charge = 50
    priority = 'high'


class RegisterTask(Task):
    name = 'register'
    n_cpus = 1
    io_charge = 10
    priority = 'high'


class CompressVideoTask(Task):
    name = 'compress_avi'
    input_path_templates = ('raw_video_data/*.avi',)
    output_path_templates = ('raw_video_data/*.mp4',)
    n_cpus = 4
    io_charge = 50
    priority = 'normal'


class RawEphysQCTask(Task):
    name = 'raw_ephys_qc'
    input_path_templates = ('raw_ephys_data/probe*/*.ap.bin',)
    output_path_templates = (
        'raw_ephys_data/probe*/_spikeglx_ephysQcTime',
        'raw_ephys_data/probe*/_spikeglx_ephysQcFreq')
    n_cpus = 4
    priority = 'normal'


class ExtractEphysTask(Task):
    name = 'extract_ephys'
    input_path_templates = ('raw_ephys_data/probe*/*.ap.bin',)
    output_path_templates = ('alf/_ibl_trials.*', 'raw_ephys_data/probe*/_spikeglx_sync*')
    n_cpus = 4
    io_charge = 50
    priority = 'high'


class SpikeSortingTask(Task):
    name = 'spike_sorting'
    input_path_templates = ('raw_ephys_data/probe*/*.ap.bin',)
    output_path_templates = ('raw_ephys_data/probe*/*.npy',)
    n_gpus = 1
    priority = 'high'


class CompressEphysTask(Task):
    name = 'compress_ephys'
    input_path_templates = ('raw_ephys_data/probe*/*.ap.bin',)
    output_path_templates = ('raw_ephys_data/probe*/*.ap.cbin',)
    n_cpus = 1
    priority = 'low'


class CompressVideoEphysTask(Task):
    name = 'compress_ephys_avi'
    input_path_templates = ()
    output_path_templates = ('raw_video_data/*.mp4',)
    n_cpus = 4
    io_charge = 50
    priority = 'normal'


class DLCTask(Task):
    name = 'dlc'
    input_path_templates = ('raw_video_data/*.mp4',)
    output_path_templates = ('alf/_dlc_*',)
    n_gpus = 1
    priority = 'normal'
    depends_on = ('compress_avi', 'compress_ephys_avi')


class SpikeSortingQCTask(Task):
    name = 'qc_spike_sorting'
    n_cpus = 1
    priority = 'high'
    depends_on = ('spike_sorting',)


class SpikeSortingMerge(Task):
    name = 'merge_spike_sorting'
    n_cpus = 1
    priority = 'high'


TASK_CLASSES = {
    'compress_avi': CompressVideoTask,
    'compress_ephys': CompressEphysTask,
    'compress_ephys_avi': CompressVideoEphysTask,
    'dlc': DLCTask,
    'extract_behaviour': ExtractBehaviorTask,
    'extract_ephys': ExtractEphysTask,
    'merge_spike_sorting': SpikeSortingMerge,
    'qc_spike_sorting': SpikeSortingQCTask,
    'raw_ephys_qc': RawEphysQCTask,
    'register': RegisterTask,
    'spike_sorting': SpikeSortingTask,
}


# -------------------------------------------------------------------------------------------------
# Pipeline routines
# -------------------------------------------------------------------------------------------------

def missing_tasks(session_dir):
    """Create a list of tasks in a session directory, by scanning the directory and
    finding which tasks have not run yet and thus have not yet created the output files."""
    for task_cls in TASK_CLASSES.values():
        task = task_cls(session_dir)
        if task.status != 'completed':
            yield task


def run_task(name, session_dir, dependencies=()):
    task = TASK_CLASSES[name](session_dir)
    task.started = datetime.now()
    logger.info("Starting task %s on %s.", name, session_dir)
    task.run()
    logger.info("Finished task %s on %s.", name, session_dir)
    task.status = 'completed'
    task.completed = datetime.now()
    return 0


def _count(l):
    return (l)


class Pipeline:
    def __init__(self, root_dir, png_path=None):
        self.session_dirs = list(find_session_dirs((root_dir,)))
        d = {}
        for session_dir in self.session_dirs:
            session_dir = str(session_dir)
            for task in missing_tasks(session_dir):
                d[('task_name', task.name)] = task.name
                dependencies = [(dt, session_dir) for dt in task.depends_on]
                d[(task.name, session_dir)] = (
                    run_task, ('task_name', task.name), session_dir, dependencies)
            d[('end', session_dir)] = (
                _count, [(task_name, session_dir) for task_name in TASK_CLASSES.keys()])
        if png_path:
            visualize(d, filename=png_path)
        self.graph = d
        self.create_cluster()

    def create_cluster(self):
        self.cluster = LocalCluster(
            n_workers=1, processes=False, silence_logs=logging.DEBUG)
        self.client = Client(self.cluster)

    def run(self):
        # TODO: check priority io etc
        # TODO: continuous server of dashboard
        return self.client.get(
            self.graph, [('end', session_dir) for session_dir in self.session_dirs])


# -------------------------------------------------------------------------------------------------
# Command-line interface
# -------------------------------------------------------------------------------------------------

if __name__ == '__main__':
    p = Pipeline('.')
    p.run()



import tempfile
from pathlib import Path, PureWindowsPath
import shutil

import numpy as np

import oneibl.webclient
import ibllib.io.raw_data_loaders as raw
from ibllib.io.extractors.biased_trials import (
    get_choice,
    get_contrastLR,
    get_feedbackType,
    get_goCueTrigger_times,
    get_intervals,
    get_iti_duration,
    get_response_times,
    get_rewardVolume,
)


def get_probabilityLeft(session_path, save=False, data=False, settings=False):
    if not data:
        data = raw.load_data(session_path)
    if not settings:
        settings = raw.load_settings(session_path)
    if settings is None:
        settings = {"IBLRIG_VERSION_TAG": "100.0.0"}
    elif settings["IBLRIG_VERSION_TAG"] == "":
        settings.update({"IBLRIG_VERSION_TAG": "100.0.0"})
    # LEN_BLOCKS is None up to v6.3.1 (also POSITIONS, CONTRASTS, et al.)
    if settings["LEN_BLOCKS"] is None:
        # Get from iblrig repo
        num = settings.get("PRELOADED_SESSION_NUM", None)
        if num is None:
            num = settings.get("PREGENERATED_SESSION_NUM", None)
        if num is None:
            fn = settings.get('SESSION_LOADED_FILE_PATH', None)
            fn = PureWindowsPath(fn).name
            num = ''.join([d for d in fn if d.isdigit()])
            if num == '':
                raise ValueError("Can't extract left probability behaviour.")
        master_branch = "https://raw.githubusercontent.com/int-brain-lab/iblrig/master/"
        sessions_folder = "tasks/_iblrig_tasks_ephysChoiceWorld/sessions/"
        fnames = [f"session_{num}_ephys_len_blocks.npy", f"session_{num}_ephys_pcqs.npy"]
        tmp = tempfile.mkdtemp()
        for fil in fnames:
            oneibl.webclient.http_download_file(f"{master_branch}{sessions_folder}{fil}",
                                                cache_dir=tmp)
        # Load the files
        len_blocks = np.load(f"{tmp}/{fnames[0]}").tolist()
        pcqs = np.load(f"{tmp}/{fnames[1]}")
        positions = [x[0] for x in pcqs]
        # Patch the settings file
        settings["LEN_BLOCKS"] = len_blocks
        settings["POSITIONS"] = positions
        # Cleanup
        shutil.rmtree(tmp)
    # Continue
    pLeft = []
    prev = 0
    for bl, nt in zip(np.cumsum(settings["LEN_BLOCKS"]), settings["LEN_BLOCKS"]):
        prob = np.sum([1 for x in np.sign(settings["POSITIONS"][prev:bl]) if x < 0]) / nt
        pLeft.extend([prob] * nt)
        prev = bl
    # Trim to actual number of trials
    pLeft = pLeft[: len(data)]
    if raw.save_bool(save, "_ibl_trials.probabilityLeft.npy"):
        lpath = Path(session_path).joinpath("alf", "_ibl_trials.probabilityLeft.npy")
        np.save(lpath, pLeft)
    return pLeft


def extract_all(session_path, save=False, data=False, output_path=None):
    """
    Extract all behaviour data from Bpod whithin the specified folder.
    The timing information from FPGA is extracted in
    :func:`~ibllib.io.extractors.ephys_fpga`

    :param session_path: folder containing sessions
    :type session_path: str or pathlib.Path
    :param save: bool
    :param data: raw Bpod data dictionary
    :return: dictionary of trial related vectors (one row per trial)
    """
    if not data:
        data = raw.load_data(session_path)
    feedbackType = get_feedbackType(session_path, save=save, data=data)
    contrastLeft, contrastRight = get_contrastLR(session_path, save=save, data=data)
    probabilityLeft = get_probabilityLeft(session_path, save=save, data=data)
    choice = get_choice(session_path, save=save, data=data)
    rewardVolume = get_rewardVolume(session_path, save=save, data=data)
    iti_dur = get_iti_duration(session_path, save=save, data=data)
    # all files containing times need to be saved with a specific bpod tag later
    intervals = get_intervals(session_path, save=False, data=data)
    response_times = get_response_times(session_path, save=False, data=data)
    go_cue_trig_times = get_goCueTrigger_times(session_path, save=False, data=data)

    out = {
        "feedbackType": feedbackType,
        "contrastLeft": contrastLeft,
        "contrastRight": contrastRight,
        "probabilityLeft": probabilityLeft,
        "session_path": session_path,
        "choice": choice,
        "rewardVolume": rewardVolume,
        "iti_dur": iti_dur,
        "goCueTrigger_times": go_cue_trig_times,
        "intervals": intervals,
        "response_times": response_times,
    }
    if save:
        output_path = Path(output_path or Path(session_path).joinpath("alf"))
        np.save(output_path.joinpath("_ibl_trials.intervals_bpod.npy"), intervals)
        np.save(output_path.joinpath("_ibl_trials.response_times_bpod.npy"), response_times)
        np.save(output_path.joinpath("_ibl_trials.goCueTrigger_times_bpod.npy"), go_cue_trig_times)
    return out


if __name__ == "__main__":
    session_path = str(Path().cwd() / "../../../tests/ibllib/extractors/data/session_ephys")
    settings = raw.load_settings(session_path)
    settings.update({"LEN_BLOCKS": None})
    data = raw.load_data(session_path)
    pLeft = get_probabilityLeft(session_path, data=data, settings=settings)
    print(".")



"""
Set of functions to handle wheel data
"""
import numpy as np
from numpy import pi
import scipy.interpolate as interpolate
from scipy.signal import convolve, gaussian
from scipy.linalg import hankel
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection


# Define some constants
ENC_RES = 1024*4  # Rotary encoder resolution, assumes X4 encoding
WHEEL_DIAMETER = 3.1*2  # Wheel diameter in cm


def interpolate_position(re_ts, re_pos, freq=1000, kind='linear', fill_gaps=None):
    """
    Return linearly interpolated wheel position.
    :param re_ts: numpy array of timestamps
    :param re_pos: numpy array of unwrapped wheel positions
    :param freq: frequency in Hz of the interpolation
    :param kind: type of interpolation; 'linear' (default) or 'cubic'
    :param fill_gaps: for gaps over this time (seconds), forward fill values before interpolation
    :return: interpolated position and timestamps
    """
    t = np.arange(re_ts[0], re_ts[-1], 1 / freq)  # Evenly resample at frequency
    yinterp = interpolate.interp1d(re_ts, re_pos, kind=kind)(t)

    if fill_gaps:
        #  Find large gaps and forward fill @fixme This is inefficient
        gaps, = np.where(np.diff(re_ts) >= fill_gaps)

        for i in gaps:
            yinterp[(t >= re_ts[i]) & (t < re_ts[i+1])] = re_pos[i]

    return yinterp, t


def velocity(re_ts, re_pos):
    """
    Compute wheel velocity from non-uniformly sampled wheel data. Returns the velocity
    at the same samples locations as the position through interpolation.

    :param re_ts: numpy array of timestamps
    :param re_pos: numpy array of unwrapped wheel positions
    :return: numpy array of velocities
    """
    dp = np.diff(re_pos)
    dt = np.diff(re_ts)
    # Compute raw velocity
    vel = dp / dt
    # Compute velocity time scale
    tv = re_ts[:-1] + dt / 2
    # interpolate over original time scale
    if tv.size > 1:
        ifcn = interpolate.interp1d(tv, vel, fill_value="extrapolate")
        return ifcn(re_ts)


def velocity_smoothed(pos, freq, smooth_size=0.03):
    """
    Compute wheel velocity from uniformly sampled wheel data
    :param pos: numpy array of wheel positions
    :param smooth_size: size of Gaussian smoothing window in seconds
    :param freq: sampling frequency of the data
    :return: tuple of velocity and acceleration values as numpy arrays
    """
    # Define our smoothing window with an area of 1 so the units won't be changed
    stdSamps = np.round(smooth_size * freq)  # Standard deviation relative to sampling frequency
    N = stdSamps * 6  # Number of points in the Gaussian
    gauss_std = (N-1) / 6  # @fixme magic number everywhere!
    win = gaussian(N, gauss_std)
    win = win / win.sum()  # Normalize amplitude

    # Convolve and multiply by sampling frequency to restore original units
    vel = np.insert(convolve(np.diff(pos), win, mode='same'), 0, 0) * freq
    acc = np.insert(convolve(np.diff(vel), win, mode='same'), 0, 0) * freq

    return vel, acc


def last_movement_onset(t, vel, event_time):
    """
    Find the time at which movement started, given an event timestamp that accured during the
    movement.  Movement start is defined as the first sample after the velocity has been zero
    for at least 50ms
    :param t: numpy array of wheel timestamps in seconds
    :param vel: numpy array of wheel velocities
    :param event_time: timestamp anywhere during movement of interest, e.g. peak velocity
    :return: timestamp of movement onset
    """
    # Look back from timestamp
    threshold = 50e-3
    mask = t < event_time
    times = t[mask]
    vel = vel[mask]
    t = None  # Initialize
    for i, t in enumerate(times[::-1]):
        i = times.size - i
        idx = np.min(np.where((t - times) < threshold))
        if np.max(np.abs(vel[idx:i])) < 0.5:
            break

    # Return timestamp
    return t


def movements(t, pos, freq=1000, pos_thresh=8, t_thresh=.2, min_gap=.1, pos_thresh_onset=1.5,
              min_dur=.05, make_plots=False):
    """
    Return linearly interpolated wheel position.
    :param t: An array of wheel timestamps in absolute seconds
    :param pos: An array of evenly sampled wheel positions
    :param freq: The sampling rate for linear interpolation
    :param pos_thresh: The minimum required movement during the t_thresh window to be considered
    part of a movement
    :param t_thresh: The time window over which to check whether the pos_thresh has been crossed
    :param min_gap: The minimum time between one movement's offset and another movement's onset
    in order to be considered separate.  Movements with a gap smaller than this are 'stictched
    together'
    :param pos_thresh_onset: A lower threshold for finding precise onset times.  The first
    position of each movement transition that is this much bigger than the starting position is
    considered the onset
    :param min_dur: The minimum duration of a valid movement.  Detected movements shorter than
    this are ignored
    :param make_plots: plot trace of position and velocity, showing detected onsets and offsets
    :return: Tuple of onset and offset times, movement amplitudes and peak velocity times
    """
    # Wheel position must be evenly sampled.
    dt = np.diff(t)
    assert np.all(np.abs(dt - dt.mean()) < 1e-10), 'Values not evenly sampled'

    # Convert the time threshold into number of samples given the sampling frequency
    t_thresh_samps = int(np.round(t_thresh * freq))
    max_disp = np.empty(t.size, dtype=float)  # initialize array of total wheel displacement

    # Calculate a Hankel matrix of size t_thresh_samps in batches.  This is effectively a
    # sliding window within which we look for changes in position greater than pos_thresh
    BATCH_SIZE = 10000  # do this in batches in order to keep memory usage reasonable
    c = 0  # index of 'window' position
    while True:
        i2proc = np.arange(BATCH_SIZE) + c
        i2proc = i2proc[i2proc < t.size]
        w2e = hankel(pos[i2proc], np.full(t_thresh_samps, np.nan))
        # Below is the total change in position for each window
        max_disp[i2proc] = np.nanmax(w2e, axis=1) - np.nanmin(w2e, axis=1)
        c += BATCH_SIZE - t_thresh_samps
        if i2proc[-1] == t.size-1:
            break

    moving = max_disp > pos_thresh  # for each window is the change in position greater than
    # our threshold?
    moving = np.insert(moving, 0, False)  # First sample should always be not moving to ensure
    # we have an onset
    moving[-1] = False  # Likewise, ensure we always end on an offset

    onset_samps = np.where(~moving[:-1] & moving[1:])[0]
    offset_samps = np.where(moving[:-1] & ~moving[1:])[0]
    too_short = np.where((onset_samps[1:] - offset_samps[:-1]) / freq < min_gap)[0]
    for p in too_short:
        moving[offset_samps[p]:onset_samps[p + 1] + 1] = True

    onset_samps = np.where(~moving[:-1] & moving[1:])[0]
    onsets_disp_arr = np.empty((onset_samps.size, t_thresh_samps))
    c = 0
    cwt = 0
    while onset_samps.size != 0:
        i2proc = np.arange(BATCH_SIZE) + c
        icomm = np.intersect1d(i2proc[:-t_thresh_samps - 1], onset_samps, assume_unique=True)
        itpltz = np.intersect1d(i2proc[:-t_thresh_samps - 1], onset_samps,
                                return_indices=True, assume_unique=True)[1]
        i2proc = i2proc[i2proc < t.size]
        if icomm.size > 0:
            w2e = hankel(pos[i2proc], np.full(t_thresh_samps, np.nan))
            w2e = np.abs((w2e.T - w2e[:, 0]).T)
            onsets_disp_arr[cwt + np.arange(icomm.size), :] = w2e[itpltz, :]
            cwt += icomm.size
        c += BATCH_SIZE - t_thresh_samps
        if i2proc[-1] >= onset_samps[-1]:
            break

    has_onset = onsets_disp_arr > pos_thresh_onset
    b, a = np.where(~np.fliplr(has_onset).T)
    # A = np.asarray([np.min(b[a == i]) for i in np.unique(a)])
    # A = np.vectorize(lambda i: np.min(b[a == i]))(np.unique(a))
    first = (np.min(b[a == i]) for i in np.unique(a))
    A = np.fromiter(first, dtype=int, count=np.unique(a).size)
    onset_lags = t_thresh_samps - A
    onset_samps = onset_samps + onset_lags - 1
    onsets = t[onset_samps]
    offset_samps = np.where(moving[:-1] & ~moving[1:])[0]
    offsets = t[offset_samps]

    durations = offsets - onsets
    too_short = durations < min_dur
    onset_samps = onset_samps[~too_short]
    onsets = onsets[~too_short]
    offset_samps = offset_samps[~too_short]
    offsets = offsets[~too_short]

    moveGaps = onsets[1:] - offsets[:-1]
    gap_too_small = moveGaps < min_gap
    if onsets.size > 0:
        onsets = onsets[np.insert(~gap_too_small, 0, True)]  # always keep first onset
        onset_samps = onset_samps[np.insert(~gap_too_small, 0, True)]
        offsets = offsets[np.append(~gap_too_small, True)]  # always keep last offset
        offset_samps = offset_samps[np.append(~gap_too_small, True)]

    move_amps = pos[offset_samps] - pos[onset_samps]
    N = 10  # Number of points in the Gaussian
    STDEV = 1.8  # Equivalent to a width factor (alpha value) of 2.5
    gauss = gaussian(N, STDEV)  # A 10-point Gaussian window of a given s.d.
    vel = convolve(np.diff(np.insert(pos, 0, 0)), gauss, mode='same')
    # For each movement period, find the timestamp where the absolute velocity was greatest
    peaks = (t[m + np.abs(vel[m:n]).argmax()] for m, n in zip(onset_samps, offset_samps))
    peak_vel_times = np.fromiter(peaks, dtype=float, count=onsets.size)

    if make_plots:
        fig, axes = plt.subplots(nrows=2, sharex='all')
        indicies = np.sort(np.hstack((onset_samps, offset_samps)))  # Points to split trace
        vel, acc = velocity_smoothed(pos, freq, 0.015)

        # Plot the wheel position and velocity
        for ax, y in zip(axes, (pos, vel)):
            ax.plot(onsets, y[onset_samps], 'go')
            ax.plot(offsets, y[offset_samps], 'bo')

            t_split = np.split(np.vstack((t, y)).T, indicies, axis=0)
            ax.add_collection(LineCollection(t_split[1::2], colors='r'))  # Moving
            ax.add_collection(LineCollection(t_split[0::2], colors='k'))  # Not moving

        axes[0].set_ylabel('position')
        axes[1].set_ylabel('velocity')
        axes[1].set_xlabel('time')
        axes[0].legend(['onsets', 'offsets', 'in movement'])
        plt.show()

    return onsets, offsets, move_amps, peak_vel_times


def cm_to_deg(positions, wheel_diameter=WHEEL_DIAMETER):
    """
    Convert wheel position to degrees turned.  This may be useful for e.g. calculating velocity
    in revolutions per second
    :param positions: array of wheel positions in cm
    :param wheel_diameter: the diameter of the wheel in cm
    :return: array of wheel positions in degrees turned
    """
    return positions / (wheel_diameter*pi) * 360


def cm_to_rad(positions, wheel_diameter=WHEEL_DIAMETER):
    """
    Convert wheel position to radians.  This may be useful for e.g. calculating angular velocity.
    :param positions: array of wheel positions in cm
    :param wheel_diameter: the diameter of the wheel in cm
    :return: array of wheel angle in radians
    """
    return positions * (2/wheel_diameter)


def samples_to_cm(positions, wheel_diameter=WHEEL_DIAMETER, resolution=ENC_RES):
    """
    Convert wheel position samples to cm linear displacement.  This may be useful for
    inter-converting threshold units
    :param positions: array of wheel positions in sample counts
    :param wheel_diameter: the diameter of the wheel in cm
    :param resolution: resolution of the rotary encoder
    :return: array of wheel angle in radians
    """
    return positions / resolution * pi * wheel_diameter


def traces_by_trial(t, pos, trials, start='stimOn_times', end='feedback_times'):
    """
    Returns list of tuples of positions and velocity for samples between stimulus onset and
    feedback.
    :param t: numpy array of timestamps
    :param pos: numpy array of wheel positions (could also be velocities or accelerations)
    :param trials: dict of trials ALFs
    :param start: trails key to use as the start index for splitting
    :param end: trails key to use as the end index for splitting
    :return: list of traces between each start and end event
    """
    if start == 'intervals' and (end is None or end == 'interval'):
        start = trials['intervals'][:, 0]
        end = trials['intervals'][:, 1]
    traces = np.vstack((pos, t))
    def to_mask(a, b): return (t > a) & (t < b)
    #  to_dict = lambda a, b, c: position: a
    cuts = [traces[:, to_mask(s, e)] for s, e in zip(trials[start], trials[end])]
    # ans = map(to_mask, zip(trials[start], trials[end]))
    # for s, e in zip(trials[start], trials[end]):
    #     mask = (wheel['times'] > s) & (wheel['times'] < e)
    #     cuts = traces[:, mask]

    return [(cuts[n][0, :], cuts[n][1, :], cuts[n][2, :]) for n in range(len(cuts))]



from pathlib import Path
import unittest
import numpy as np
import pickle
import brainbox.behavior.wheel as wheel


class TestWheel(unittest.TestCase):

    def setUp(self):
        # Test data is in the form ((inputs), (outputs)) where inputs is a tuple containing a
        # numpy array of timestamps and one of positions; outputs is a tuple of outputs from
        # teh function under test, e.g. wheel.movements
        pickle_file = Path(__file__).parent.joinpath('wheel_test.pickle')
        if not pickle_file.exists():
            self.test_data = None
        else:
            with open(pickle_file, 'rb') as f:
                self.test_data = pickle.load(f)

    def test_derivative(self):
        if self.test_data is None:
            return
        t = np.array([0, .5, 1., 1.5, 2, 3, 4, 4.5, 5, 5.5])
        p = np.arange(len(t))
        v = wheel.velocity(t, p)
        self.assertTrue(len(v) == len(t))
        self.assertTrue(np.all(v[0:4] == 2) and v[5] == 1 and np.all(v[7:] == 2))
        # import matplotlib.pyplot as plt
        # plt.figure()
        # plt.plot(t[:-1] + np.diff(t) / 2, np.diff(p) / np.diff(t), '*-')
        # plt.plot(t, v, '-*')

    def test_movements(self):
        if self.test_data is None:
            return
        # These test data are the same as those used in the MATLAB code
        inputs = self.test_data[0][0]
        expected = self.test_data[0][1]
        on, off, amp, peak_vel = wheel.movements(
            *inputs, freq=1000, pos_thresh=8, pos_thresh_onset=1.5)
        self.assertTrue(np.array_equal(on, expected[0]), msg='Unexpected onsets')
        self.assertTrue(np.array_equal(off, expected[1]), msg='Unexpected offsets')
        self.assertTrue(np.array_equal(amp, expected[2]), msg='Unexpected move amps')
        # Differences due to convolution algorithm
        all_close = np.allclose(peak_vel, expected[3], atol=1.e-2)  
        self.assertTrue(all_close, msg='Unexpected peak velocities')

    def test_movements_FPGA(self):
        if self.test_data is None:
            return
        # These test data are the same as those used in the MATLAB code.  Test data are from 
        # extracted FPGA wheel data
        pos, t = wheel.interpolate_position(*self.test_data[1][0], freq=1000)
        expected = self.test_data[1][1]
        thresholds = wheel.samples_to_cm(np.array([8, 1.5]))
        on, off, amp, peak_vel = wheel.movements(
            t, pos, freq=1000, pos_thresh=thresholds[0], pos_thresh_onset=thresholds[1])
        self.assertTrue(np.allclose(on, expected[0], atol=1.e-5), msg='Unexpected onsets')
        self.assertTrue(np.allclose(off, expected[1], atol=1.e-5), msg='Unexpected offsets')
        self.assertTrue(np.allclose(amp, expected[2], atol=1.e-5), msg='Unexpected move amps')
        self.assertTrue(np.allclose(peak_vel, expected[3], atol=1.e-2), 
                        msg='Unexpected peak velocities')
